from typing import Any

import numpy
from scipy.stats import poisson

from .PropertyGenerator import PropertyGenerator


class PoissonNumberGenerator(PropertyGenerator):
    def __init__(self, mu: float, return_int: bool = False):
        """
        Init a NumberGenerator which will output number taken from a skewed normal distribution.

        :param mu: average number of events per interval.
        :param return_int: return number as integer instead of float
        """
        self.mu = mu
        self.return_int = return_int

    def generate(self) -> Any:
        generated = poisson.rvs(self.mu)
        if self.return_int:
            return numpy.asscalar(generated.round())
        else:
            return numpy.asscalar(generated)



from typing import Any

from .PropertyGenerator import PropertyGenerator


class FixedValueGenerator(PropertyGenerator):
    def __init__(self, value: Any):
        """
        Init a FixedValueGenerator which will always generate the same value. Used for non-changing property.

        :param value: value to output
        """
        self.value = value

    def generate(self) -> Any:
        return self.value



import csv
import random
from typing import Any, Dict, List, Optional

from .EventGenerator import EventGenerator


class EventComposer:
    def __init__(self):
        """
        Init an EventComposer used to aggregate multiple event generators
        """
        self.samples: Dict[int, int] = {}
        self.generators: Dict[int, EventGenerator] = {}

    def add_generator(self, event_generator: EventGenerator, samples: int):
        """
        Add an event generator to this composer

        :param event_generator: event generator to add to this composer
        :param samples: number of samples to draw from this generator
        """
        self.generators[id(event_generator)] = event_generator
        self.samples[id(event_generator)] = samples

    def remove_generator(self, generator: EventGenerator):
        """
        Remove a event generator from this composer

        :param generator: event generator to remove from this composer
        """
        del self.generators[id(generator)]
        del self.samples[id(generator)]

    def generate(self, shuffle: bool = True) -> List[Dict[str, Any]]:
        """
        Generate all events from this composer

        :param shuffle: Shuffle events in the resulting list. False by default.
        :return: a list of generated events
        """
        total_events = []

        for _id, generator in self.generators.items():
            samples = self.samples[_id]
            events = [generator.generate() for _ in range(samples)]
            total_events.extend(events)

        if shuffle:
            random.shuffle(total_events)
        return total_events

    def to_csv(self, destination: str, shuffle: bool = False, properties: Optional[List[str]] = None):
        """
        Generate all events from this composer and save them as csv

        :param destination: path for outputted csv
        :param shuffle: Shuffle events in the resulting list. False by default.
        :param properties: List of properties used to order columns in CSV. If not set, columns will have random order.
        """
        events = self.generate(shuffle)
        if properties is None:
            properties = set([key for generator in self.generators.values() for key in generator.keys()])
        with open(destination, 'w') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=properties)

            writer.writeheader()
            writer.writerows(events)



from typing import Any, Dict, List

from .EventGenerator import EventGenerator
from ..property.PropertyGenerator import PropertyGenerator


class PropertyEventGenerator(EventGenerator):

    def __init__(self, properties: Dict[str, PropertyGenerator]):
        """
        Init an Event Generator based on a DateGenerator and multiples PropertyGenerator

        :param properties: a dict like {property_name: PropertyGenerator}
        """
        self.properties = properties

    def generate(self) -> Dict[str, Any]:
        generated_properties = {name: prop.generate() for name, prop in self.properties.items()}
        return {
            **generated_properties
            }

    def keys(self)-> List[str]:
        return [*list(self.properties.keys())]



from abc import ABC, abstractmethod
from typing import Any


class PropertyGenerator(ABC):
    @abstractmethod
    def generate(self) -> Any:
        """
        Generate a random property

        :return: a random property
        """
        raise NotImplementedError



from .DateGenerator import DateGenerator
from .FixedValueGenerator import FixedValueGenerator
from .NormalNumberGenerator import NormalNumberGenerator, NormalNumberGenerator as NumberGenerator
from .RandomChoiceGenerator import RandomChoiceGenerator



from datetime import datetime

from scipy.stats import skewnorm

from mojito.property.PropertyGenerator import PropertyGenerator


class DateGenerator(PropertyGenerator):
    def __init__(self, center: datetime, skew: float = 0, deviation: float = 1):
        """
        Create a DateGenerator. This generator is based on a second basis.

        It means that if you ask a date around 2018-01-01T00:00:00 with a scale of 10, you will get date +- 5s around
        this date.

        :param skew: Skew parameters. Number below 0 will result in more date to the left limit.
        :param center: Mean generated datetime for the skewed normal distribution
        :param deviation: Deviation of datetime, in second, used for the skewed normal distribution
        """
        self.skew = skew
        self.center = center
        self.dispersion = deviation

    def generate(self) -> datetime:
        """
        Generate a random date taken from a Skewed Normal Distribution

        :return: datetime
        """
        timestamp = round(self.center.timestamp())
        generated = skewnorm.rvs(self.skew, loc=timestamp, scale=self.dispersion)
        return datetime.fromtimestamp(generated)



from setuptools import setup, find_packages

with open("Readme.md") as f:
    long_desc = f.read()

setup(
        name='mojito-mock',
        version='0.3.0',
        packages=find_packages(),
        license='',
        author='Pierre Paci',
        author_email='',
        description='Mocking temporal data made easy !',
        long_description=long_desc,
        url="https://github.com/PPACI/mojito",
        long_description_content_type="text/markdown",
        classifiers=(
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: Apache Software License",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering"
        ),
        install_requires=[
            "numpy==1.13.3",
            "scipy==1.1.0"
            ]
        )



# Expose event generator
from .event.PropertyEventGenerator import PropertyEventGenerator
from .event.EventComposer import EventComposer

# Expose property generator
from .property import DateGenerator, FixedValueGenerator, NormalNumberGenerator, NumberGenerator, RandomChoiceGenerator




from abc import ABC, abstractmethod
from typing import Any, Dict, List


class EventGenerator(ABC):
    @abstractmethod
    def generate(self) -> Dict[str, Any]:
        """
        Generate a random event

        :return: a dict like {property_name: value}
        """
        raise NotImplementedError

    @abstractmethod
    def keys(self) -> List[str]:
        """
        Return the list of properties name contained in this event generator

        :return: a list of properties name
        """
        raise NotImplementedError



import random
from typing import Any, List, Optional

import numpy

from .PropertyGenerator import PropertyGenerator


class RandomChoiceGenerator(PropertyGenerator):
    def __init__(self, initial_list: List[Any], weights: Optional[List[float]] = None):
        """
        Init a RandomChoiceGenerator which will output a random choice in the given list.

        :param initial_list: list to randomly take value
        :param weights: Optional weights list used to alter the probabilities
        """
        self.initial_list = initial_list
        if weights is not None:
            assert len(weights) == len(initial_list), "Len of weights and initial_list should be the same"
        else:
            weights = numpy.ones(len(initial_list))
        self.weights = weights

    def generate(self) -> Any:
        choices = random.choices(self.initial_list, weights=self.weights)
        return choices[0]



from typing import Any

import numpy
from scipy.stats import skewnorm

from .PropertyGenerator import PropertyGenerator


class NormalNumberGenerator(PropertyGenerator):
    def __init__(self, mean: float, deviation: float, skew: float = 0, return_int: bool = False):
        """
        Init a NumberGenerator which will output number taken from a skewed normal distribution.

        :param skew: Skew parameters. Number below 0 will skew distribution to the left.
                    Set to 0 to have a standard Normal Number Generator.
        :param mean: Mean generated number
        :param deviation: deviation of the distribution
        :param return_int: return number as integer instead of float
        """
        self.skew = skew
        self.mean = mean
        self.deviation = deviation
        self.return_int = return_int

    def generate(self) -> Any:
        generated = skewnorm.rvs(self.skew, loc=self.mean, scale=self.deviation)
        if self.return_int:
            return numpy.asscalar(generated.round())
        else:
            return numpy.asscalar(generated)
